//----------------------------------------------------------------------------
//      _____
//     *     *
//    *____   *____
//   * *===*   *==*
//  *___*===*___**  AVNET
//       *======*
//        *====*
//----------------------------------------------------------------------------
//
// This design is the property of Avnet.  Publication of this
// design is not authorized without written consent from Avnet.
//
// Please direct any questions to the BCM4343W IoT Starter Kit community support forum:
//    http://cloudconnectkits.org/forum
//
// Product information is available at:
//    http://cloudconnectkits.org/product/avnet-bcm4343w-iot-starter-kit
//    http://cloudconnectkits.org/product/avnet-bcm4343w-soc-module
//
// Disclaimer:
//    Avnet, Inc. makes no warranty for the use of this code or design.
//    This code is provided  "As Is". Avnet, Inc assumes no responsibility for
//    any errors, which may appear in this code, nor does it make a commitment
//    to update the information contained herein. Avnet, Inc specifically
//    disclaims any implied warranties of fitness for a particular purpose.
//                     Copyright(c) 2016 Avnet, Inc.
//                             All rights reserved.
//
//----------------------------------------------------------------------------

/*
 * ZentriOS SDK LICENSE AGREEMENT | Zentri.com, 2016.
 *
 * Use of source code and/or libraries contained in the ZentriOS SDK is
 * subject to the Zentri Operating System SDK license agreement and
 * applicable open source license agreements.
 *
 */
/* Documentation for this app is available online.
 * See https://docs.zentri.com/wifi/sdk/latest/examples/cloud/bluemix
 */




#include "zos.h"
#include "platform_common.h"
#include "common.h"
#include "led_matrix8x8.h"

#define DISPLAY_BRIGHTNESS      15  // max 15
#define DISPLAY_SCROLL_RATE_MS  64
#define DISPLAY_BLINK_RATE_OFF  0

/** @file
 *
 * Based on the IBM Bluemix Example App using the MQTT protocol
 *
 * The application connects to the Bluemix Quickstart broker:
 * quickstart.messaging.internetofthings.ibmcloud.com
 *
 * After a connection with the broker is established, the following sensor
 * readings are published at 5 second intervals. Both readings are
 * combined into a JSON formatted message
 *   - Signal strength (RSSI) of the Wi-Fi access point the board is
 *     connected with
 *   - Light sensor value converted using ZentriOS ADC APIs
 *
 * The sensor readings can be viewed in real-time on the IBM Watson IoT
 * Platform at https://quickstart.internetofthings.ibmcloud.com
 *
 * The Device ID is zentri_XXXX where XXXX is the least significant
 * 4 bytes of the WLAN MAC address. The ID is printed to a ZentriOS terminal.
 *
 */

/******************************************************
 *                      Macros
 ******************************************************/

#if PLATFORM_ADC_MAX_RESOLUTION == 12
#define LUT_FILENAME  "lut_12bit.csv"
#else
#define LUT_FILENAME  "lut_10bit.csv"
#endif

/******************************************************
 *                    Constants
 ******************************************************/
#define BUTTON_DEBOUNCE_TIME  50 // ms
#define BUTTON_CLICK_TIME    500 // ms
#define BUTTON_PRESS_TIME    750 // ms

int periodic_flag = 0; //periodic publish is off until turned on
/******************************************************
 *                   Enumerations
 ******************************************************/

/******************************************************
 *                 Type Definitions
 ******************************************************/

/******************************************************
 *                    Structures
 ******************************************************/
const zos_adc_config_t ADC_CONFIG =
{
    .resolution = PLATFORM_ADC_MAX_RESOLUTION,
    .sampling_cycle = 12,
    .gain = 1
};

/******************************************************
 *               Function Declarations
 ******************************************************/
static zos_result_t mqtt_connection_event_cb( mqtt_event_info_t *event );

/******************************************************
 *               Variable Definitions
 ******************************************************/
mqtt_settings_t settings;
mqtt_connection_t* mqtt_connection;
static mqtt_callback_t callback = mqtt_connection_event_cb;
zos_adc_lut_t adc_lut;

/******************************************************
 *               Function Definitions
 ******************************************************/
void zn_app_init(void)
{
    ZOS_LOG("Starting 'Voice Automation Demo' App");

    // GPIO Initialization //
    zn_issue_command("gdi 17 out");
    zn_issue_command("gdi 28 out");
    zn_issue_command("gdi 27 out");
    zn_issue_command("gdi 26 out");

	// LED Initialization //
    PLATFORM_ENABLE_JTAG_GPIOS();
    zn_gpio_init(PLATFORM_LED1, GPIO_OUTPUT_PUSHPULL, 0);
    zn_gpio_init(PLATFORM_LED2, GPIO_OUTPUT_PUSHPULL, 0);

    // LED Matrix Initialization //
    led_matrix8x8_init(PLATFORM_STD_I2C);
    led_matrix8x8_set_text(" VOICE AUTOMATION DEMO"); //space makes the wrap-around look smoother
    led_matrix8x8_set_brightness(DISPLAY_BRIGHTNESS);
    led_matrix8x8_set_blink_rate(DISPLAY_BLINK_RATE_OFF);
    led_matrix8x8_set_scroll_rate(DISPLAY_SCROLL_RATE_MS);

    zos_result_t result;
    zos_adc_t adc;
    char mac_str[32] = {0};

    // Get device MAC address //
    zn_network_get_mac(mac_str);

    // Bring up the network interface //
    if (!zn_network_is_up(ZOS_WLAN))
    {
        ZOS_LOG("Network is down, restarting...");
        if(ZOS_FAILED(result, zn_network_restart(ZOS_WLAN)))
        {
            ZOS_LOG("Failed to restart network: %d", result);
            return;
        }
    }

    // ADC Initialization //
    if(ZOS_FAILED(result, zn_adc_gpio_to_peripheral(PLATFORM_STD_ADC, &adc))) //pf
    {
        ZOS_LOG("The specified GPIO: %d does not support ADC!", PLATFORM_STD_ADC);
    }

    if (ZOS_FAILED(result, zn_adc_direct_init(adc, &ADC_CONFIG)))
    {
        ZOS_LOG("Failed to initialize ADC!");
    }

    // Prepare ADC lookup table //
    if (ZOS_FAILED(result, zn_adc_add_lut(ZOS_ADC_GET_MASK(adc), LUT_FILENAME, &adc_lut)))
    {
        ZOS_LOG("Failed to prepare ADC lookup table! Error code %d", result);
    }

    // MQTT Settings Initialization //
    strcpy((char*)settings.host, MQTT_HOST);
    sprintf((char*)settings.topic, MQTT_TOPIC, mac_str[12], mac_str[13], mac_str[15], mac_str[16]);
    sprintf((char*)settings.client, MQTT_CLIENT_ID, mac_str[12], mac_str[13], mac_str[15], mac_str[16]);
    strcpy((char*)settings.user, MQTT_USER);
    strcpy((char*)settings.password, MQTT_PASSWORD);
    settings.port = MQTT_PORT;
    settings.keepalive = MQTT_KEEPALIVE;
    settings.qos = MQTT_QOS;
    settings.security = MQTT_SECURITY;

    ZOS_LOG("\nDevice ID: "DEVICE_ID"\r\n", mac_str[12], mac_str[13], mac_str[15], mac_str[16]);
    ZOS_LOG("To view published sensor data, go to: https://quickstart.internetofthings.ibmcloud.com/#/"DEVICE_ID"/", mac_str[12], mac_str[13], mac_str[15], mac_str[16]);

    // Memory allocated for MQTT object //
    zn_malloc((uint8_t**)&mqtt_connection, sizeof(mqtt_connection_t));
    if ( mqtt_connection == NULL )
    {
        ZOS_LOG("Failed to allocate MQTT object...\n");
        return;
    }

    // MQTT library Initialization //
    result = mqtt_init( mqtt_connection );
    if(result != ZOS_SUCCESS)
    {
        ZOS_LOG("Error initializing");
        zn_free(mqtt_connection);
        mqtt_connection = NULL;
        return;
    }

    ZOS_LOG("\r\nIBM Bluemix MQTT Demo Application Started");
    ZOS_LOG("  - Broker: %s", settings.host);
    ZOS_LOG("  - Client ID: %s", settings.client);
    ZOS_LOG("  - Topic/Queue: %s\r\n", settings.topic);

    // Connect to IBM broker Bluemix broker //
    mqtt_app_connect(NULL);
}


/*************************************************************************************************/
void zn_app_deinit(void)
{
    // Free MQTT object and deinit library //
    mqtt_deinit( mqtt_connection );
    zn_free(mqtt_connection);
    mqtt_connection = NULL;

    // Remove ADC lookup table //
    zn_adc_remove_lut(adc_lut);
    led_matrix8x8_deinit();
}

/*************************************************************************************************/
/*
 * Connect to broker (sent connect frame)
 */
void mqtt_app_connect( void *arg )
{
    mqtt_pkt_connect_t conninfo;
    zos_result_t ret = ZOS_SUCCESS;

    ZOS_LOG("Opening connection with broker %s:%u", settings.host, settings.port);

    ret = mqtt_open( mqtt_connection, (const char*)settings.host, settings.port, ZOS_WLAN, callback, settings.security );
    if ( ret != ZOS_SUCCESS )
    {
        ZOS_LOG("Error opening connection (keys and certificates are set properly?)");
        return;
    }

    ZOS_LOG("Connection established");

    // Now, after socket is connected we can send the CONNECT frame safely //
    //ZOS_LOG("Connecting...");
    memset( &conninfo, 0, sizeof( conninfo ) );

    conninfo.mqtt_version = MQTT_PROTOCOL_VER4;
    conninfo.clean_session = 1;
    conninfo.client_id = settings.client;
    conninfo.keep_alive = settings.keepalive;
    conninfo.password = strlen((char*)settings.password) > 0 ? settings.password : NULL;
    conninfo.username = strlen((char*)settings.user) > 0 ? settings.user : NULL;
    ret = mqtt_connect( mqtt_connection, &conninfo );
    if ( ret != ZOS_SUCCESS )
    {
        ZOS_LOG("Error connecting");
    }
    zn_event_issue(mqtt_app_subscribe, NULL, 0);

}

/*************************************************************************************************/
/*
 * Disconnect from broker (sent disconnect frame)
 */
void mqtt_app_disconnect( void *arg )
{
    if ( mqtt_disconnect( mqtt_connection ) != ZOS_SUCCESS )
    {
        ZOS_LOG("Error disconnecting");
    }
}

/*************************************************************************************************/
/*
 * Subscribe to topic
 */
void mqtt_app_subscribe( void *arg )
{
    mqtt_msgid_t pktid;
    ZOS_LOG("Subscribing to topic '%s'", settings.topic);
    pktid = mqtt_subscribe( mqtt_connection, settings.topic, settings.qos );
    if ( pktid == 0 )
    {
        ZOS_LOG("Error subscribing: packet ID = 0");
    }
}

/*************************************************************************************************/
/*
 * Unsubscribe from topic
 */
void mqtt_app_unsubscribe( void *arg )
{
    mqtt_msgid_t pktid;
    ZOS_LOG("Unsubscribing from topic '%s'", settings.topic);
    pktid = mqtt_unsubscribe( mqtt_connection, settings.topic );

    if ( pktid == 0 )
    {
        ZOS_LOG("Error unsubscribing: packet ID = 0");
    }
}

/*************************************************************************************************/
/*
 * Publish (send) message to topic
 */
void mqtt_app_publish( void *arg )
{
    mqtt_msgid_t pktid;
    ZOS_LOG("----------------------------");
    ZOS_LOG("Topic: %s, Message: '%s'", settings.topic_pub,settings.message);
    ZOS_LOG("----------------------------");
    pktid = mqtt_publish( mqtt_connection, settings.topic_pub, settings.message, (uint32_t)strlen((char*)settings.message), settings.qos );

    if ( pktid == 0 )
    {
        ZOS_LOG("Error publishing: packet ID = 0");
    }
}

/******************************************************
 *               Static Function Definitions
 ******************************************************/
/*
 * Call back function to handle connection events.
 */
static zos_result_t mqtt_connection_event_cb( mqtt_event_info_t *event )
{
    switch ( event->type )
    {
        case MQTT_EVENT_TYPE_CONNECTED:
            ZOS_LOG("CONNECTED" );
            break;
        case MQTT_EVENT_TYPE_DISCONNECTED:
            ZOS_LOG("DISCONNECTED" );
            break;
        case MQTT_EVENT_TYPE_PUBLISHED:
            ZOS_LOG("MESSAGE PUBLISHED" );
        	if (periodic_flag == 1) { // keep publishing unless commanded to stop
        	zn_event_register_timed(periodic_publish, NULL, MQTT_PUBLISH_PERIOD, 0);
        	}
            break;
        case MQTT_EVENT_TYPE_SUBCRIBED:
            ZOS_LOG("TOPIC SUBSCRIBED" );
            break;
        case MQTT_EVENT_TYPE_UNSUBSCRIBED:
            ZOS_LOG("TOPIC UNSUBSCRIBED" );
            break;
        case MQTT_EVENT_TYPE_PUBLISH_MSG_RECEIVED:
        {
        	mqtt_topic_msg_t msg = event->data.pub_recvd;

            ZOS_LOG("MESSAGE RECEIVED");
            //ZOS_LOG("Topic  : %.*s", msg.topic_len, msg.topic );
            //ZOS_LOG("Message: %.*s", msg.data_len, msg.data);

    		msg.topic[msg.topic_len] = '\0';
        	strcpy((char*)settings.topic_pub, (char*)msg.topic);
            ZOS_LOG("Topic  : %s", settings.topic_pub );
            reset_relay(NULL);

            switch(msg.data[9]) {
            	case 49: //STOP

                    ZOS_LOG("----------------------------");
            		ZOS_LOG("Message: STOP!");
                	reset_LED(NULL);
					zn_gpio_set(PLATFORM_LED2, 1);
                	reset_relay(NULL);
                	zn_issue_command("gse 17 1");
					led_matrix8x8_set_text(" STOP");
		            ZOS_LOG("----------------------------");
	            	break;

            	case 50: //GO

                    ZOS_LOG("----------------------------");
            		ZOS_LOG("Message: GO!");
                	reset_LED(NULL);
                	zn_gpio_set(PLATFORM_LED1, 1);
                	reset_relay(NULL);
                	zn_issue_command("gse 28 1");
                	led_matrix8x8_set_text(" GO");
                    ZOS_LOG("----------------------------");
                	break;

            	case 51: //WAIT

                    ZOS_LOG("----------------------------");
                	ZOS_LOG("Message: WAIT!");
                	reset_LED(NULL);
                	reset_relay(NULL);
                	zn_issue_command("gse 27 1");
                	led_matrix8x8_set_text(" WAIT");
                    ZOS_LOG("----------------------------");
            	    break;

            	case 52: //COOL

                    ZOS_LOG("----------------------------");
            		ZOS_LOG("Message: COOL!");
                	reset_LED(NULL);
                	reset_relay(NULL);
                	zn_issue_command("gse 26 1");
                	led_matrix8x8_set_text(" FAN");
                    ZOS_LOG("----------------------------");
            		break;

            	case 53: //START PUBLISH

                    ZOS_LOG("----------------------------");
            		ZOS_LOG("Message: Starting Periodic Publish...");
            		periodic_flag = 1;
            		zn_event_issue(periodic_publish, NULL, 0);
                    ZOS_LOG("----------------------------");
            		break;

            	case 54: //STOP PUBLISH

                    ZOS_LOG("----------------------------");
            		ZOS_LOG("Message: Stopping Periodic Publish...");
            		periodic_flag = 0;
                    ZOS_LOG("----------------------------");
            		break;

            	case 55: // message display

                    ZOS_LOG("----------------------------");
            		msg.data += 19;
            		msg.data[0] = ' ';
            		msg.data[strlen(msg.data)-3] = '\0';
            		ZOS_LOG("Message:%s",msg.data);
            		led_matrix8x8_set_text(msg.data);
                    ZOS_LOG("----------------------------");
            		break;

            	case 57: //RESET

                    ZOS_LOG("----------------------------");
            		ZOS_LOG("Message: Resetting...");
            		reset_relay(NULL);
            	    reset_LED(NULL);
            		led_matrix8x8_deinit();
                    ZOS_LOG("----------------------------");
            		break;

            	default:

            		break;
            }

        }
            break;
        default:
            break;
    }
    return ZOS_SUCCESS;
}
/*************************************************************************************************/
/*
 * Periodic publish the WiFi Status and Light Sensor Readings
 */

static void periodic_publish(void *arg)
{
    char light_sense_buffer[16];
    char ssid_buffer[16];
    char rssi_buffer[16];
    char channel_buffer[100];

    zos_result_t result;
    zos_adc_t  adc;
    uint16_t   adc_raw = 0, light_value = 0;

    uint32_t bytes_read1;
    uint32_t bytes_read2;
    uint32_t bytes_read3;

    // Light Sensor Status //
    if(ZOS_FAILED(result, zn_adc_gpio_to_peripheral(PLATFORM_STD_ADC, &adc)))
    {
        ZOS_LOG("The specified GPIO: %d does not support ADC!", PLATFORM_STD_ADC);
    }

    if (ZOS_FAILED(result, zn_adc_direct_sample(adc, &adc_raw)))
    {
        ZOS_LOG("Failed to read converted ADC sample! Error code %d", result);
    }
    else
    {
    light_value = (4095 - adc_raw);
    int_to_str(light_value, light_sense_buffer);

        // Publish the converted ADC sample to MQTT queue //
        sprintf((char*)settings.message, MQTT_MESSAGE_2, light_sense_buffer, 5);
        mqtt_app_publish(NULL);
    }

    // WiFi Status //
	    zn_issue_command_return_data(ssid_buffer, sizeof(ssid_buffer), &bytes_read1, "get wl s");
	    ssid_buffer[bytes_read1] = 0;
	    zn_issue_command_return_data(rssi_buffer, sizeof(rssi_buffer), &bytes_read2, "wlan_get_rssi");
	    rssi_buffer[bytes_read2] = 0;
	    zn_issue_command_return_data(channel_buffer, sizeof(channel_buffer), &bytes_read3, "get wlan.info");
	    channel_buffer[bytes_read3] = 0;

	   char sub[12];
	   int position = 0, length = 12;
	   while (strncmp(sub,"channel:",8) != 0) {
	   int c = 0;
	   position++;
	   	   while (c < length) {
	   		   sub[c] = channel_buffer[position+c-1];
	   		   c++;
	   	   }
	   }
	   if (sub[10] == '\r')
	   {
		   sprintf((char*)settings.message, MQTT_MESSAGE_0, ssid_buffer,sub[9],rssi_buffer,6);
	   }

	   else
	   {
		   sprintf((char*)settings.message, MQTT_MESSAGE_1, ssid_buffer,sub[9],sub[10],rssi_buffer,6);
	   }

        mqtt_app_publish(NULL);
}

/*************************************************************************************************/
/*
 * Reset the LED's
 */
static void reset_LED(void *arg)
{
	if (zn_gpio_get(PLATFORM_LED1) == 1)
	{
		zn_gpio_set(PLATFORM_LED1,0);
	}
	if (zn_gpio_get(PLATFORM_LED2) == 1)
	{
		zn_gpio_set(PLATFORM_LED2,0);
	}
	return;
}

/*************************************************************************************************/
/*
 * Reset the Relays
 */
static void reset_relay(void *arg)
{
    zn_issue_command("gse 17 0");
    zn_issue_command("gse 28 0");
    zn_issue_command("gse 27 0");
    zn_issue_command("gse 26 0");

    return;
}
